/*
 * Copyright (c) 2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Mon Feb 27 2017
 */
#include "acheader.h"

#include <avr/interrupt.h>
#include <avr/io.h>

#include "arduio.h"
#include <util/atomic.h>
#include <util/delay.h>


static enum ario_dir ario_dpins[14];
#ifdef ARDUIO_UNO
static uint8_t ario_apins[6];
#elif defined(ARDUIO_NANO)
static uint8_t ario_apins[8];
#else
static uint8_t ario_apins[8];
#endif

/* --- ario_pin_init ------------------------------------------------------- */

void
ario_pin_init(void)
{
  uint8_t i;

  /* timers at max frequency, fast PWM */
  TCCR0A = (1 << WGM01) | (1 << WGM00);
  TCCR0B = (0 << WGM02) | (0 << CS02) | (0 << CS01) | (1 << CS00);
  OCR0A = OCR0B = 0;

  TCCR1A = (0 << WGM11) | (1 << WGM10);
  TCCR1B = (0 << WGM13) | (1 << WGM12) |
    (0 << CS12) | (0 << CS11) | (1 << CS10);
  OCR1A = OCR1B = 0;

  TCCR2A = (1 << WGM21) | (1 << WGM20);
  TCCR2B = (0 << WGM22) | (0 << CS22) | (0 << CS21) | (1 << CS20);
  OCR2A = OCR2B = 0;

  /* ADC, AVCC ref, single conversion, 125kHz (~100µs/chan), left aligned */
  ADMUX = (0 << REFS1) | (1 << REFS0) | (1 << ADLAR);
  DIDR0 =
    (1 << ADC5D) | (1 << ADC4D) | (1 << ADC3D) |
    (1 << ADC2D) | (1 << ADC1D) | (1 << ADC0D);
  ADCSRB = (0 << ADTS2) | (0 << ADTS1) | (0 << ADTS0);
  ADCSRA =
    (1 << ADEN) | (1 << ADSC) | (0 << ADATE) | (1 << ADIF) | (1 << ADIE) |
    (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);


  /* input pins by default */
  for(i = 0; i < sizeof(ario_dpins)/sizeof(ario_dpins[0]); i++)
    ario_dpin_dir(i, ARIO_IN);
}


/* --- ario_dpin_dir ------------------------------------------------------- */

void
ario_dpin_dir(uint8_t n, enum ario_dir dir)
{
  if (n >= sizeof(ario_dpins)/sizeof(ario_dpins[0])) return;
  if (n < 2) return;

  /* configure PWM first */
  switch(dir) {
    case ARIO_IN:
    case ARIO_INPU:
    case ARIO_OUT:
      switch(n) {
        case  3: TCCR2A &= ~((1 << COM2B1) | (0 << COM2B0)); break;
        case  5: TCCR0A &= ~((1 << COM0B1) | (0 << COM0B0)); break;
        case  6: TCCR0A &= ~((1 << COM0A1) | (0 << COM0A0)); break;
        case  9: TCCR1A &= ~((1 << COM1A1) | (0 << COM1A0)); break;
        case 10: TCCR1A &= ~((1 << COM1B1) | (0 << COM1B0)); break;
        case 11: TCCR2A &= ~((1 << COM2A1) | (0 << COM2A0)); break;
      }
      break;

    case ARIO_PWM:
      switch(n) {
        case  3: TCCR2A |= (1 << COM2B1) | (0 << COM2B0); break;
        case  5: TCCR0A |= (1 << COM0B1) | (0 << COM0B0); break;
        case  6: TCCR0A |= (1 << COM0A1) | (0 << COM0A0); break;
        case  9: TCCR1A |= (1 << COM1A1) | (0 << COM1A0); break;
        case 10: TCCR1A |= (1 << COM1B1) | (0 << COM1B0); break;
        case 11: TCCR2A |= (1 << COM2A1) | (0 << COM2A0); break;

        default: return;
      }
      break;
  }

  /* configure direction */
  switch(dir) {
    case ARIO_IN:
      if (n < 8) {	/* portd pins */
        PORTD &= ~(1 << n);
        DDRD &= ~(1 << n);
      } else {		/* portb pins */
        PORTB &= ~(1 << (n-8));
        DDRB &= ~(1 << (n-8));
      }
      break;

    case ARIO_INPU:
      if (n < 8) {	/* portd pins */
        DDRD &= ~(1 << n);
        PORTD |= (1 << n);
      } else {		/* portb pins */
        DDRB &= ~(1 << (n-8));
        PORTB |= (1 << (n-8));
      }
      break;

    case ARIO_PWM:
    case ARIO_OUT:
      if (n < 8)	/* portd pins */
        DDRD |= (1 << n);
      else		/* portb pins */
        DDRB |= (1 << (n-8));
      break;
  }

  ario_dpins[n] = dir;
}


/* --- ario_dpin_out ------------------------------------------------------- */

void
ario_dpin_out(uint8_t n, uint8_t v)
{
  if (n >= sizeof(ario_dpins)/sizeof(ario_dpins[0])) return;
  if (n < 2) return;

  switch(ario_dpins[n]) {
    case ARIO_IN:
    case ARIO_INPU:
      break;

    case ARIO_OUT:
      if (n < 8)	/* portd pins */
        if (v) PORTD |= (1 << n); else PORTD &= ~(1 << n);
      else		/* portb pins */
        if (v) PORTB |= (1 << (n-8)); else PORTB &= ~(1 << (n-8));
      break;

    case ARIO_PWM:
      switch(n) {
        case  3: OCR2B = v; break;
        case  5: OCR0B = v; break;
        case  6: OCR0A = v; break;
        case  9: OCR1A = v; break;
        case 10: OCR1B = v; break;
        case 11: OCR2A = v; break;

        default: return;
      }
      break;
  }
}


/* --- ario_dpin_in -------------------------------------------------------- */

uint8_t
ario_dpin_in(uint8_t n)
{
  if (n >= sizeof(ario_dpins)/sizeof(ario_dpins[0])) return 0;
  if (n < 2) return 0;

  switch(ario_dpins[n]) {
    case ARIO_PWM:
      switch(n) {
        case  3: return OCR2B;
        case  5: return OCR0B;
        case  6: return OCR0A;
        case  9: return OCR1A;
        case 10: return OCR1B;
        case 11: return OCR2A;
      }
      /*FALLTHROUGH*/
    case ARIO_IN:
    case ARIO_INPU:
    case ARIO_OUT:
      if (n < 8)	/* portd pins */
        return (PIND & (1 << n)) ? 0xff : 0;
      else		/* portb pins */
        return (PINB & (1 << (n-8))) ? 0xff : 0;
      break;
  }

  return 0;
}


/* --- ario_apin_in -------------------------------------------------------- */

uint8_t
ario_apin_in(uint8_t n)
{
  if (n >= sizeof(ario_apins)/sizeof(ario_apins[0])) return 0;
  return ario_apins[n];
}


/* --- ario_ddir_state ----------------------------------------------------- */

void
ario_ddir_state(void)
{
  uint8_t dir[1 + sizeof(ario_dpins)/sizeof(ario_dpins[0])];
  uint8_t i;

  dir[0] = 'D';
  for(i = 0; i < sizeof(ario_dpins)/sizeof(ario_dpins[0]); i++)
    dir[i+1] = ario_dpins[i];

  ario_log_buffer(dir, sizeof(dir)/sizeof(dir[0]));
}


/* --- ario_dpin_state ----------------------------------------------------- */

void
ario_dpin_state(void)
{
  uint8_t in[1 + sizeof(ario_dpins)/sizeof(ario_dpins[0])];
  uint8_t i;

  in[0] = 'I';
  for(i = 0; i < sizeof(ario_dpins)/sizeof(ario_dpins[0]); i++)
    in[i+1] = ario_dpin_in(i);

  ario_log_buffer(in, sizeof(in)/sizeof(in[0]));
}


/* --- ario_apin_state ----------------------------------------------------- */

void
ario_apin_state(void)
{
  uint8_t in[1 + sizeof(ario_apins)/sizeof(ario_apins[0])];
  uint8_t i;

  in[0] = 'A';
  for(i = 0; i < sizeof(ario_apins)/sizeof(ario_apins[0]); i++)
    in[i+1] = ario_apins[i];

  ario_log_buffer(in, sizeof(in)/sizeof(in[0]));
}


/* --- ADC_vect ------------------------------------------------------------ */

ISR(ADC_vect)
{
  uint8_t channel = ADMUX & 0x7;

  ario_apins[channel++] = ADCH;
  if (channel >= sizeof(ario_apins)/sizeof(ario_apins[0]))
    channel = 0;
  ADMUX = (ADMUX & 0xe0) | (channel & 0x7);

  NONATOMIC_BLOCK(NONATOMIC_FORCEOFF) {
    _delay_us(10); /* delay a bit to stabilize input */
  }
  ADCSRA |= (1 << ADSC);
}
